@@ -1,8 +1,8 @@
 #ifndef HLFL__BSD_IPFW_H
-#define HLHL__BSD_IPFW_H
+#define HLFL__BSD_IPFW_H
 
 int translate_bsd_ipfw(int, char *, char *, int, char *, char *, char *,
 		       char *);
 int translate_bsd_ipfw_start();
 void include_text_ipfw();
-#endif				/* HLHL__BSD_IPFW_H */
+#endif				/* HLFL__BSD_IPFW_H */
