@@ -1,6 +1,6 @@
 /* vi:set tw=80 sts=1 ts=1 sw=1:
  * hlfl
- * Copyright (C) 2000 Renaud Deraison
+ * Copyright (C) 2000-2002 Renaud Deraison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -61,6 +61,7 @@ char *error_str[] = {
 };
 
 int matched_if = 0;
+char *lang = 0;
 
 translator_start_t translator_start;
 translate_t translate_func;
@@ -104,8 +105,10 @@ get_definition(d)
 
  while (d[0] == ' ')
   d++;
+
  while (d[strlen(d) - 1] == ' ')
   d[strlen(d) - 1] = '\0';
+
  if ((t = strchr(d, ' ')))
   t[0] = '\0';
 
@@ -147,7 +150,6 @@ remove_spaces(t)
  if (!t)
   return t;
 
-
  while ((t[0] == ' ') || (t[0] == '\t'))
   t++;
  while ((t[strlen(t) - 1] == ' ') || (t[strlen(t) - 1] == '\t'))
@@ -155,6 +157,22 @@ remove_spaces(t)
  return t;
 }
 
+char *
+next_op(op)
+ char *op;
+{
+ if (!op)
+  return NULL;
+
+ op = strchr(op, ' ');
+ if (op)
+   {
+    while (op[0] == ' ')
+     op++;
+   }
+ return op;
+}
+
 /*
  * Returns the integer value of the operator
  */
@@ -162,34 +180,150 @@ int
 int_op(op)
  char *op;
 {
- if (!strcmp(op, "->"))
-  return ACCEPT_ONE_WAY;
- if (!strcmp(op, "<-"))
-  return ACCEPT_ONE_WAY_REVERSE;
- if (!strcmp(op, "<->"))
-  return ACCEPT_TWO_WAYS;
- if (!strcmp(op, "<=>>"))
-  return ACCEPT_TWO_WAYS_ESTABLISHED;
- if (!strcmp(op, "<<=>"))
-  return ACCEPT_TWO_WAYS_ESTABLISHED_REVERSE;
- if (!strcmp(op, "X"))
-  return DENY;
- if (!strcmp(op, "X!"))
-  return REJECT;
- if (!strcmp(op, "X->"))
-  return DENY_OUT;
- if (!strcmp(op, "<-X"))
-  return DENY_IN;
- if (!strcmp(op, "X!->") || !strcmp(op, "!X->"))
-  return REJECT_OUT;
- if (!strcmp(op, "<-X!"))
-  return REJECT_IN;
+ int ret = 0;
+
+ /*
+  * * FIXME op tests need to be ordered
+  * * complex before, simple next
+  */
+
+ while (op && strlen(op))
+   {
+    if (!strncmp(op, "<->", strlen("<->")))
+      {
+       ret |= ACCEPT_TWO_WAYS;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "<=>>", strlen("<=>>")))
+      {
+       ret |= ACCEPT_TWO_WAYS_ESTABLISHED;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "<<=>", strlen("<<=>")))
+      {
+       ret |= ACCEPT_TWO_WAYS_ESTABLISHED_REVERSE;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "X->", strlen("X->")))
+      {
+       ret |= DENY_OUT;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "<-X", strlen("<-X")))
+      {
+       ret |= DENY_IN;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "X!->", strlen("X!->")) ||
+	     !strncmp(op, "!X->", strlen("!X->")))
+      {
+       ret |= REJECT_OUT;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "<-X!", strlen("<-X!")))
+      {
+       ret |= REJECT_IN;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "X!", strlen("X!")))
+      {
+       ret |= REJECT_ALL;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "X", strlen("X")))
+      {
+       ret |= DENY_ALL;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "->", strlen("->")))
+      {
+       ret |= ACCEPT_ONE_WAY;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "<-", strlen("<-")))
+      {
+       ret |= ACCEPT_ONE_WAY_REVERSE;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "accept", strlen("accept")))
+      {
+       ret |= ACCEPT;
+       op = next_op(op);
+      }
+
+    else if (!strncmp(op, "deny", strlen("deny")))
+      {
+       ret |= DENY;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "reject", strlen("reject")))
+      {
+       ret |= REJECT;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "to", strlen("to")))
+      {
+       ret |= ONE_WAY;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "from", strlen("from")))
+      {
+       ret |= ONE_WAY_REVERSE;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "established", strlen("established")))
+      {
+       ret |= ESTABLISHED;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "log", strlen("log")))
+      {
+       ret |= LOG;
+       op = next_op(op);
+      }
+    else if (!strncmp(op, "and", strlen("and")))
+      {
+       op = next_op(op);
+      }
+    else
+      {
+       error = HLFL_SYNTAX_ERROR;
+       return -1;
+      }
+   }
+
+ if (!ret)
+   {
+    error = HLFL_UNKNOWN_OP;
+    return -1;			/* error */
+   }
+
+ /*
+  * Sanity checks
+  */
+ if (((ret & (ACCEPT | DENY | REJECT)) != ACCEPT) &&
+     ((ret & (ACCEPT | DENY | REJECT)) != DENY) &&
+     ((ret & (ACCEPT | DENY | REJECT)) != REJECT))
+   {
+    error = HLFL_SYNTAX_ERROR;
+    return -1;
+   }
+
+ if ((ret == ACCEPT) || (ret == DENY) || (ret == REJECT))
+  ret |= TWO_WAYS;
+
+ if ((ret | LOG) == (ACCEPT | ESTABLISHED | TWO_WAYS | LOG))
+   {
+    /*
+     * XXX fixme !
+     */
+    ret -= ESTABLISHED;
+   }
+
+ return ret;
 
- error = HLFL_UNKNOWN_OP;
- return -1;			/* error */
 }
 
-
 int
 icmp(proto)
  char *proto;
@@ -197,7 +331,6 @@ icmp(proto)
  return !strcmp(proto, "icmp");
 }
 
-
 /*
  * returns TRUE if proto in {all, tcp, udp, icmp)
  */
@@ -223,38 +356,38 @@ translate_proto(proto)
  int current = 0;
 
  ret = malloc((MAX_PROTOS + 1) * sizeof(char *));
- bzero(ret, sizeof(char *) * (MAX_PROTOS + 1));
+ memset(ret, 0, sizeof(char *) * (MAX_PROTOS + 1));
 
  t = proto;
-if (s = strchr(t, '|'))
-{
- while (s)
+ if ((s = strchr(t, '|')))
+   {
+    while (s)
+      {
+       s[0] = '\0';
+       if (check_proto(remove_spaces(t)))
+	 {
+	  error = HLFL_UNKNOWN_PROTOCOL;
+	  return NULL;
+	 }
+       if (current + 1 >= MAX_PROTOS)
+	 {
+	  error = HLFL_TOO_MANY_PROTOCOLS;
+	  return NULL;
+	 }
+       ret[current++] = strdup(t);
+       t = s + 1;
+       s[0] = '|';
+       s = strchr(t, '|');
+      }
+   }
+ else
    {
-    s[0] = '\0';
     if (check_proto(remove_spaces(t)))
       {
        error = HLFL_UNKNOWN_PROTOCOL;
        return NULL;
       }
-    if (current + 1 >= MAX_PROTOS)
-      {
-       error = HLFL_TOO_MANY_PROTOCOLS;
-       return NULL;
-      }
-    ret[current++] = strdup(t);
-    t = s + 1;
-    s[0] = '|';
-    s = strchr(t, '|');
    }
-}
-else
-{
- if (check_proto(remove_spaces(t)))
-  {
-   error = HLFL_UNKNOWN_PROTOCOL;
-   return NULL;
-  }
-}
  ret[current++] = strdup(t);
  ret[current] = NULL;
  return ret;
@@ -272,9 +405,8 @@ length(s)
  return r - 1;
 }
 
-
 /*
- * Returns 1 if the IP adress <s> is valid (of the form x.y.z.t)
+ * Returns nonzero if the IP adress <s> is valid (of the form x.y.z.t)
  */
 int
 valid_ip(s)
@@ -300,7 +432,6 @@ valid_ip(s)
  return ok;
 }
 
-
 /*
  * Convert : "ip ports |  ip ports | ip ports" to
  * an array of strings
@@ -332,7 +463,7 @@ ip(src, level)
  t = remove_spaces(t);
 
  ret = malloc(n * sizeof(char *));
- bzero(ret, n * sizeof(char *));
+ memset(ret, 0, n * sizeof(char *));
 
  while (t)
    {
@@ -340,11 +471,9 @@ ip(src, level)
     if (s)
      s[0] = '\0';
 
-
     if (!valid_ip(t))
       {
 
-
        /*
         * 't' is not a valid ip (x.y.z.t)
         *
@@ -366,11 +495,12 @@ ip(src, level)
 	  w[0] = '\0';
 	  end = w + 1;
 	  r = ip(t, level + 1);
-	  if(!r)
-	  {
-	   if(!error)error = HLFL_UNDEF_VAR_ERROR;
-	   return NULL;
-	  }
+	  if (!r)
+	    {
+	     if (!error)
+	      error = HLFL_UNDEF_VAR_ERROR;
+	     return NULL;
+	    }
 	  while (r[i])
 	    {
 	     ret[current] = malloc(strlen(end) + strlen(r[i]) + 1);
@@ -457,7 +587,7 @@ ifaces(iface, level)
  t = remove_spaces(iface);
 
  ret = malloc(n * sizeof(char *));
- bzero(ret, n * sizeof(char *));
+ memset(ret, 0, n * sizeof(char *));
 
  t = iface;
  while (t)
@@ -508,11 +638,10 @@ ifaces(iface, level)
  * The function that calls the appropriate translator...
  */
 int
-translate(proto, src, op, log, dst, interface, flags)
+translate(proto, src, op, dst, interface, flags)
  char *proto;
  char *src;
  char *op;
- int log;
  char *dst;
  char *interface;
  char *flags;
@@ -525,18 +654,21 @@ translate(proto, src, op, log, dst, interface, flags)
  char **protos;
  char **interfaces = NULL;
  int ni = 0;
-
-
+ int log = 0;
 
  if ((iface = get_definition(interface)))
    {
     interface = iface;
    }
 
-
  if (opi < 0)
   return opi;
 
+ if (opi & LOG)
+   {
+    log++;
+    opi -= LOG;
+   }
  if (!(protos = translate_proto(proto)))
    {
     error = HLFL_UNKNOWN_PROTOCOL;
@@ -556,7 +688,6 @@ translate(proto, src, op, log, dst, interface, flags)
        srcs = ip(src, 0);
        dsts = ip(dst, 0);
 
-
        if (!srcs || !dsts)
 	 {
 	  if (!error)
@@ -680,7 +811,6 @@ translate(proto, src, op, log, dst, interface, flags)
  return 0;
 }
 
-
 /*---------------------------------------------------------------------
 
 			       (ugly) Parser
@@ -742,7 +872,6 @@ include(file)
  return 0;
 }
 
-
 /*
  * Define a variable
  *
@@ -777,7 +906,6 @@ define(order)
  return 0;
 }
 
-
 /*
  * Process one single line
  */
@@ -791,17 +919,14 @@ process(buffer)
  char *src;
  char *dst;
  char *op;
- int log = 0;
- char *interface;
+ char *interface = NULL;
  char *flags;
  char old_t;
  int n;
 
-
  while ((t[0] == ' ') || (t[0] == '\t') || (t[0] == '\n'))
   t++;
 
-
  /*
   * Drop the comments
   */
@@ -814,14 +939,12 @@ process(buffer)
  if (t[0] == '!')
   return INCLUDE_TEXT;
 
-
  /*
   * Our syntax is :
   *
   *     proto (src) op (dst) [interface] flags
   */
 
-
  /* proto */
  proto = t;
  t = strchr(proto + 1, ' ');
@@ -855,7 +978,6 @@ process(buffer)
     while (t[0] == ' ')
      t++;
 
-
     /* src */
     if (t[0] != '(')
       {
@@ -884,32 +1006,17 @@ process(buffer)
     /* op */
     op = t;
 
-    t = strchr(op + 1, ' ');
+    t = strchr(op + 1, '(');
     if (!t)
       {
-       t = strchr(op + 1, '(');
-       if (!t)
-	 {
-	  error = HLFL_SYNTAX_ERROR;
-	  return -1;
-	 }
+       error = HLFL_SYNTAX_ERROR;
+       return -1;
       }
     old_t = t[0];
     t[0] = '\0';
     op = strdup(op);
     t[0] = old_t;
 
-    /* log */
-
-    if (strchr(op, 'l'))
-      {
-       char * x;
-       x = strchr(op, 'l');
-       memcpy(x, x+1, strlen(x+1));
-       op[strlen(op)-1]='\0';
-       log = 1;
-      }
-
     while ((t[0] == ' ') || (t[0] == '\t'))
      t++;
 
@@ -954,23 +1061,68 @@ process(buffer)
        while (t[0] == ' ')
 	t++;
       }
-    else
-     interface = NULL;
-
+    else if (!strncmp(t, "on", 2))
+      {
+       t += 2;
+       while (t[0] == ' ')
+	t++;
+       if (t[0] == '(')
+	 {
+	  interface = t + 1;
+	  if (!(t = matching_items(t, '(', ')')))
+	    {
+	     error = HLFL_SYNTAX_ERROR;
+	     return -1;
+	    }
+	  t[0] = '\0';
+	  interface = strdup(interface);
+	  t++;
+	  while (t[0] == ' ')
+	   t++;
+	 }
+       else if (t[0] == '[')
+	 {
+	  interface = t + 1;
+	  if (!(t = matching_items(t, '[', ']')))
+	    {
+	     error = HLFL_SYNTAX_ERROR;
+	     return -1;
+	    }
+	  t[0] = '\0';
+	  interface = strdup(interface);
+	  t++;
+	  while (t[0] == ' ')
+	   t++;
+	 }
+       else
+	 {
+	  interface = t;
+	  t = strchr(t + 1, ' ');
+	  if (t)
+	   t[0] = '\0';
+	  interface = strdup(interface);
+	  if (interface[strlen(interface) - 1] == '\n')
+	   interface[strlen(interface) - 1] = '\0';
+	  if (t)
+	    {
+	     t++;
+	     while (t[0] && t[0] == ' ')
+	      t++;
+	    }
+	 }
+      }
     /* extra flags (optional) */
 
-    if (t[0] && (t[0] != '\n'))
+    if (t && t[0] && (t[0] != '\n'))
       {
        flags = strdup(t);
       }
     else
      flags = NULL;
 
-
     n = translate(remove_spaces(proto),
 		  remove_spaces(src),
 		  remove_spaces(op),
-		  log,
 		  remove_spaces(dst),
 		  remove_spaces(interface), remove_spaces(flags));
     free(proto);
@@ -990,7 +1142,6 @@ print_comment(buffer)
  printf(buffer);
 }
 
-
 void
 read_file(file, fname)
  FILE *file;
@@ -998,7 +1149,7 @@ read_file(file, fname)
 {
  char buffer[4096];
  int line = 0;
- bzero(buffer, sizeof(buffer));
+ memset(buffer, 0, sizeof(buffer));
  while (fgets(buffer, sizeof(buffer) - 1, file))
    {
     int n;
@@ -1022,16 +1173,16 @@ read_file(file, fname)
        t++;			/* t[0] == '!' */
        while (t[0] == '\n' || t[0] == '\t' || t[0] == ' ')
 	t++;
-	
-       if(!strncmp(t, "else", strlen("else")))
-        {
-	 if(!matched_if)
-          printf("%s", t+strlen("else"));
-	}
+
+       if (!strncmp(t, "else", strlen("else")))
+	 {
+	  if (!matched_if)
+	   printf("%s", t + strlen("else"));
+	 }
        else
-        include_text_func(t);
+	include_text_func(t);
       }
-    bzero(buffer, sizeof(buffer));
+    memset(buffer, 0, sizeof(buffer));
    }
 }
 
@@ -1047,16 +1198,17 @@ usage(n)
 {
 
  fprintf(stderr, "%s version %s\n", n, VERSION);
- fprintf(stderr, "Copyright (C) 2000 Renaud Deraison");
- fprintf(stderr, "< deraison @ cvs.nessus.org > \n\n");
+ fprintf(stderr, "Copyright (C) 2000-2002 Renaud Deraison");
+ fprintf(stderr, " < deraison @ hlfl.org > \n\n");
  fprintf(stderr, "Usage : %s type < input > output\n", n);
  fprintf(stderr, "Where <type> can be one of :\n");
  fprintf(stderr, "\tipfw - BSD ipfw\n");
+ fprintf(stderr, "\tipfw4 - BSD ipfw 4 (stateful)\n");
  fprintf(stderr, "\tipfilter - Darren Reeds's ipfilter\n");
  fprintf(stderr, "\tipfwadm - Linux 2.0.x ipfwadm\n");
  fprintf(stderr, "\tipchains - Linux 2.2.x ipchains\n");
- fprintf(stderr, "\tnetfilter - Linux 2.3.x netfilter\n");
- fprintf(stderr, "\tcisco - Cisco rules\n\n");
+ fprintf(stderr, "\tnetfilter - Linux 2.4.x netfilter\n");
+ fprintf(stderr, "\tcisco - Cisco rules (IOS 12.1(2)T)\n\n");
  exit(1);
 }
 
@@ -1066,13 +1218,12 @@ main(argc, argv)
  char **argv;
 {
 
-
  if (!argv[1] || argc != 2)
    {
     usage(argv[0]);
    }
 
- if (!strcmp(argv[1], "ipfw"))
+ if (!strcmp(argv[1], "ipfw") || !strcmp(argv[1], "ipfw4"))
    {
     translator_start = translate_bsd_ipfw_start;
     translate_func = translate_bsd_ipfw;
@@ -1130,6 +1281,7 @@ main(argc, argv)
  /*
   * Always define the symbol 'any'
   */
+ lang = argv[1];
  add_definition("any", "0.0.0.0/0");
  translator_start();
  read_file(stdin, "stdin");
